from django.db import models
from django.contrib.contenttypes.models import ContentType
from bloomerp.models.core import BloomerpModel, ApplicationField
from bloomerp.models.fields import CodeField, TextEditorField


# ---------------------------------
# Document Template Model
# ---------------------------------
class DocumentTemplateHeader(BloomerpModel):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template_header'

    name = models.CharField(max_length=100, blank=False, null=False) #Name of the document template header
    header = models.ImageField() #Content of the header
    
    def __str__(self):
        return self.name
    
# ---------------------------------
# Document Template Free Variable Model
# ---------------------------------
class DocumentTemplateFreeVariable(BloomerpModel):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template_free_variable'

    VARIABLE_TYPE_CHOICES = [
        ('date', 'Date'),
        ('boolean', 'Boolean'),
        ('text', 'Text'),
        ('list', 'List'),
        ('integer', 'Integer'),
        ('float', 'Decimal'),
        ('model','Model')
    ]
    
    name = models.CharField(max_length=100, blank=False, null=False) #Name of the free variable
    variable_type = models.CharField(max_length=10, choices=VARIABLE_TYPE_CHOICES, blank=False, null=False)
    options = models.TextField(null=True, blank=True)
    required = models.BooleanField(null=False, blank=False, default=False)

    @property
    def slug(self):
        return self.name.replace(' ','_').lower()
    
    def __str__(self):
        return self.name
    

# ---------------------------------
# Document Template Styling Model
# ---------------------------------
class DocumentTemplateStyling(BloomerpModel):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template_styling'

    name = models.CharField(max_length=100, blank=False, null=False) #Name of the document template styling
    styling = CodeField(language='css', default='') #Content of the styling
    
    def __str__(self):
        return self.name


# ---------------------------------
# Document Template Model
# ---------------------------------
class DocumentTemplate(BloomerpModel):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template'

    name = models.CharField(max_length=100, blank=False, null=False) #Name of the document template
    template = TextEditorField(default='') # Content of the template, including the variables
    model_variable = models.ForeignKey(ContentType, on_delete=models.SET_NULL, null=True, blank=True) # Many to many field to Content Type
    free_variables = models.ManyToManyField(DocumentTemplateFreeVariable,blank=True) # Many to many field of free variable, a free variable is a variable that is not from a model
    template_header = models.ForeignKey(DocumentTemplateHeader, on_delete=models.SET_NULL, null=True, blank=True) #Foreign key to the document template header
    footer = models.TextField(null=True,blank=True)
    standard_document = models.BooleanField(null=False,blank=False, default=False) # Signifies weather the document should be created on object creation
    styling = models.ForeignKey(DocumentTemplateStyling, on_delete=models.SET_NULL, null=True, blank=True) # Foreign key to the document template styling

    def __str__(self):
        return self.name

    allow_string_search = True
    string_search_fields = ['name']

    def get_related_content_types(model):
        related_content_types = [ContentType.objects.get_for_model(model)]
        return related_content_types

    @property
    def application_fields(self):
        '''
        Returns a queryset of ApplicationField that are related to the model variable of the document template.
        '''
        qs = ApplicationField.objects.filter(content_type=self.model_variable)
        return qs

    @staticmethod
    def get_standard_documents_for_instance(instance):
        content_type = ContentType.objects.get_for_model(instance)
        return DocumentTemplate.objects.filter(model_variable=content_type, standard_document=True)
    



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest, FileResponse
from bloomerp.utils.document_templates import DocumentController
from bloomerp.models import DocumentTemplate
from django.shortcuts import get_object_or_404
from django.views.decorators.clickjacking import xframe_options_exempt
import json
from django.contrib.auth.decorators import login_required

@login_required
@xframe_options_exempt
@route('preview_document_template')
def preview_document_template(request:HttpRequest) -> HttpResponse:
    # Some permissions check
    if not request.user.has_perm('bloomerp.view_documenttemplate'):
        return HttpResponse('User does not have permission to view document templates')
    
    try:
        data : dict = json.loads(request.body)
    except json.JSONDecodeError:
        return HttpResponse(status=400)
    
    template_id = data.get('template_id')
    content = data.get('content')

    document_template = get_object_or_404(DocumentTemplate, id=template_id)
    if content:
        document_template.template = content

    try:
        document_controller = DocumentController()
        file_bytes = document_controller.create_preview_document(
            template=document_template,
            data={}
        )

        # Return the PDF file as a FileResponse
        response = HttpResponse(file_bytes, content_type='application/pdf')
        response['Content-Disposition'] = 'inline; filename="dynamic.pdf"'
        return response
    except Exception as e:
        # Display errors to the user in the template

        return HttpResponse(f'An error occured in the template: {e}', status=200)



from django.db import models
import ast
from bloomerp.models.core import File
from django.core.exceptions import ValidationError
import os

# ---------------------------------
# Bloomerp List Field
# ---------------------------------
class ListField(models.TextField):
    description = "Stores a python list"

    def __init__(self, *args, **kwargs):
        super(ListField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        # If the value is empty or None, return an empty list
        if not value:
            return []
        # If the value is already a list, just return it
        if isinstance(value, list):
            return value
        # Otherwise, parse the string to a list using ast.literal_eval
        return ast.literal_eval(value)

    def get_prep_value(self, value):
        # If the value is None, return None
        if value is None:
            return value
        # Convert the value to a string (previously 'unicode' in Python 2)
        return str(value)

    def value_to_string(self, obj):
        # Get the field's value from the object
        value = self.value_from_object(obj)
        # Convert the value to a format suitable for storage in the database
        return self.get_prep_value(value)


# ---------------------------------
# Bloomerp File Field
# ---------------------------------
class BloomerpFileField(models.ForeignKey):
    def __init__(self, *args, allowed_extensions=None, **kwargs):
        """
        Initialize BloomerpFileField with a ForeignKey to the File model and optional file type validation.
        `allowed_extensions` specifies allowed file types; if '__all__' or None, all are allowed.
        """
        self.allowed_extensions = allowed_extensions if allowed_extensions is not None else '__all__'
        kwargs['to'] = 'bloomerp.File'
        kwargs['on_delete'] = models.SET_NULL
        
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        """
        Specifies the default form field and widget to use with this model field.
        """
        from bloomerp.widgets.bloomerp_file_field_widget import BloomerpFileFieldWidget

        defaults = {
            'widget': BloomerpFileFieldWidget(),
        }

        defaults.update(kwargs)
        return super().formfield(**defaults)
        
    def validate_file_extension(self, file_instance:File):
        """
        Validate the file extension of the file associated with the foreign key.
        """
        # If allowed_extensions is '__all__', no restriction on file types
        if self.allowed_extensions == '__all__':
            return

        # Get the file extension of the associated file
        ext = os.path.splitext(file_instance.file.name)[1].lower()

        # Check if the extension is in the allowed list
        if ext not in self.allowed_extensions:
            allowed_ext_str = ', '.join(self.allowed_extensions)
            raise ValidationError(f'Unsupported file extension. Allowed extensions are: {allowed_ext_str}')

    def clean(self, value, model_instance):
        """
        Perform the validation on the foreign key reference and ensure the file type is allowed.
        """
        # Call the parent clean method to validate the ForeignKey relationship
        value = super().clean(value, model_instance)

        file_instance = File.objects.get(pk=value)

        # Validate the file extension of the linked file object
        if value:  # Ensure that a valid file instance is passed
            self.validate_file_extension(file_instance)

        return value

# ---------------------------------
# Bloomerp Code Field
# ---------------------------------
class CodeField(models.TextField):
    '''
    A custom model field to store code snippets with syntax highlighting.
    '''
    def __init__(self, *args, language='python', **kwargs):
        self.language = language
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        """
        This method tells Django how to serialize the field for migrations.
        """
        name, path, args, kwargs = super().deconstruct()
        kwargs['language'] = self.language
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        """
        Specifies the default form field and widget to use with this model field.
        """
        from django import forms
        from bloomerp.widgets.code_editor_widget import AceEditorWidget  # Import your custom widget

        defaults = {
            'form_class': forms.CharField,
            'widget': AceEditorWidget(language=self.language),
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)

    # Optional: Add custom validation logic if needed
    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        # Example: Add syntax validation for Python code
        if self.language == 'python':
            import ast
            try:
                ast.parse(value)
            except SyntaxError as e:
                raise ValidationError(f"Invalid Python code: {e}")


# ---------------------------------
# Bloomerp Text Editor Field
# ---------------------------------
class TextEditorField(models.TextField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def formfield(self, **kwargs):
        from django import forms
        from bloomerp.widgets.text_editor import RichTextEditorWidget

        defaults = {
            'form_class': forms.CharField,
            'widget': RichTextEditorWidget(),
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)



# Generated by Django 5.1.1 on 2024-09-15 22:17

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0005_alter_order_options_alter_product_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="order",
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "permissions": (
                    (
                        "bulk_delete%(class)s",
                        "Can delete multiple objects from %(class)s at once",
                    ),
                    (
                        "bulk_change%(class)s",
                        "Can update multiple objects from %(class)s at once",
                    ),
                )
            },
        ),
    ]



from bloomerp.models import (
    DocumentTemplate,
    File)
from bloomerp.utils.pdf import generate_pdf
from django.db.models import Model
from django.core.files.base import ContentFile
from django.template import engines
from django.template.loader import render_to_string

class DocumentController:
    def __init__(self) -> None:
        pass

    def create_document(
            self,
            template:DocumentTemplate, 
            instance:Model, 
            free_variables: dict=None
        ):
        ''' Creates a document for a particular template, using the model variable and the free variables:
            - Template : The document template
            - Instance : The model instance
            - Free variables : A dictionary of free variables

        '''
        data = {}

        data['object'] = instance
    	
        #Add free variable data
        data.update(free_variables)

        #Create metadata variable
        meta_data = {}
        meta_data['document_template'] = template.pk

        meta_data['signed'] = False
        
        #Format HTML       
        django_engine = engines["django"]
        temp = django_engine.from_string(template.template)
        formatted_html = temp.render(data)        

        document_bytes = generate_pdf(
            html_content=formatted_html,
            css_content=template.styling.styling,
        )
    
        content_file = ContentFile(document_bytes)
        
        file_object = File()
        
        #Save the file
        file_object.file.save(template.name + ' ' + str(instance) + '.pdf', content_file)

        file_object.name = template.name + ' ' + str(instance)
        file_object.content_object = instance

        #Save metadata
        file_object.meta = meta_data

        file_object.save()
        
    def create_preview_document(
            self,
            template:DocumentTemplate,
            data:dict
        ) -> ContentFile:
        ''' Creates a preview document for a particular template, using the model variable and the free variables:
            - Template : The document template
            - Data : A dictionary of free variables
        '''
        #Format HTML
        django_engine = engines["django"]
        temp = django_engine.from_string(template.template)
        formatted_html = temp.render(data)
        
        document_bytes = generate_pdf(
            html_content=formatted_html,
            css_content=template.styling.styling,
        )

        content_file = ContentFile(document_bytes)

        return content_file



    def send_signature_request(
            self,
            requested_by:int,
            file_id: int,
            requested_for = None,
            requested_for_email = None,
        ):
        '''
        Sends a signature request to someone.
        '''
        #Get file
        file = File.objects.get(pk=int(file_id))
        '''
        if requested_for_email:
            #Email given
            signature_request = SignatureRequest(
                requested_by=User.objects.get(pk=int(requested_by)),
                requested_for_email=requested_for_email,
                file = file
            )
            email = requested_for_email
        else:
            #User given
            requested_for_user = User.objects.get(pk=int(requested_for))

            signature_request = SignatureRequest(
                requested_by=User.objects.get(pk=int(requested_by)),
                requested_for=requested_for_user,
                file = File.objects.get(pk=int(file_id))
            )
            email = requested_for_user.email
        signature_request.save()
        
        #Send notification
        NotificationController.create_notification_from_signature_request(signature_request)
        '''

    def sign_pdf(self, pdf_file : File, signature_bytes):
        '''
        Function that will sign a pdf file, using signature bytes.
        '''
        file_path = pdf_file.file.path

        meta_data = {
            'signed' : True
        }
        '''
        handler = PdfHandler(file_path)

        #Sign the actual document and retreive the bytes
        document_bytes = handler.sign_pdf(
            signature_path=None,
            signature_bytes= signature_bytes
        )
        content_file = ContentFile(document_bytes)
        
        signed_file_obj = File()

        #Save the file
        signed_file_obj.file.save(pdf_file.name + '- signed' '.pdf', content_file)

        signed_file_obj.name = pdf_file.name + ' - signed'
        signed_file_obj.content_object = pdf_file.content_object
        
        signed_file_obj.meta = {'signed':True}
        signed_file_obj.save()
        
        pdf_file.meta = {'signed_file_id':signed_file_obj.pk}
        pdf_file.save()

        return signed_file_obj
        '''

        
         


        


        


from django.db import models

# Create your models here.
class Parent(models.Model):
    name = models.CharField(max_length=255)
    age = models.IntegerField()
    is_active = models.BooleanField(default=True)


class Child(models.Model):
    name = models.CharField(max_length=255)
    age = models.IntegerField()
    is_active = models.BooleanField(default=True)
    date = models.DateField()
    parent = models.ForeignKey(Parent, on_delete=models.CASCADE)




from typing import Any
from django.forms import ValidationError
from bloomerp.models import ApplicationField, User, File, UserDetailViewTab, Link, UserListViewPreference
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.forms.models import modelform_factory
from django.utils import timezone
from datetime import timedelta
from django.core.files import File as DjangoFile
from django.template.loader import get_template
from django.db.models import Model
from uuid import UUID
from bloomerp.utils.models import (
    get_file_fields_dict_for_model,
    get_bloomerp_file_fields_for_model,
    get_foreign_key_fields_for_model
    )
from django.contrib.postgres.fields import JSONField
from django.db.models import JSONField as DefaultJSONField
# ---------------------------------
# Bloomerp Bulk Upload Form
# ---------------------------------
class BulkUploadForm(forms.Form):

    def __init__(self, *args, **kwargs):
        super(BulkUploadForm, self).__init__(*args, **kwargs)
        for name, field in self.fields.items():
            if isinstance(field, forms.BooleanField):
                self.fields[name].widget = forms.Select(choices=[(True, 'True'), (False, 'False')])

        # Add delete all objects
        self.fields['delete_all'] = forms.BooleanField(required=False, label='Delete all selected objects',initial=False)

        # Remove last_updated_by field
        if 'last_updated_by' in self.fields:
            del self.fields['last_updated_by']

# ---------------------------------
# Object file form
# ---------------------------------

class ObjectFileForm(forms.ModelForm):
    class Meta:
        model = File
        fields = ['name', 'file']

    def __init__(self,
                 related_object:Model=None,
                 user:User=None,
                *args, **kwargs):

        self.related_object = related_object
        self.user = user

        super().__init__(*args, **kwargs)


    def save(self, commit=True):
        # Override the save method to set the content_type and object_id
        instance:File = super().save(commit=False)

        if self.related_object:
            instance.content_type = ContentType.objects.get_for_model(self.related_object)
            instance.object_id = self.related_object.pk
    
        if self.user:
            instance.uploaded_by = self.user


        if commit:
            instance.save()
        return instance

# ---------------------------------
# Bloomerp Model Form
# ---------------------------------
from bloomerp.widgets.foreign_key_widget import ForeignKeyWidget
from bloomerp.widgets.code_editor_widget import AceEditorWidget
from bloomerp.widgets.multiple_model_select_widget import MultipleModelSelect
from django.forms.widgets import DateInput, DateTimeInput


class BloomerpModelForm(forms.ModelForm):
    model:Model = None
    user:User = None
    instance:Model = None
    is_new_instance:bool = True

    def __init__(self, model:Model, user:User=None, *args, **kwargs):
        
        # Set the model instance to the form instance
        self.model = model
        self._meta.model = model
        self.user = user

        super(BloomerpModelForm, self).__init__(*args, **kwargs)

        # Set the instance to the form instance
        instance:Model = kwargs.get('instance')
        if instance:
            self.instance = instance
            self.is_new_instance = False

        # Get all of the foreign key fields for the model
        self.foreign_key_fields = get_foreign_key_fields_for_model(self.model)


        # ---------------------------------
        # FOREIGN KEY FIELDS
        # ---------------------------------
        # Update the widgets for the foreign key fields
        for field in self.foreign_key_fields:
            # Get the related model
            if field.field in self.fields:
                related_model = field.meta['related_model']
                model = ContentType.objects.get(pk=related_model).model_class()
                self.fields[field.field].widget = ForeignKeyWidget(model=model)
        
        # ---------------------------------
        # MANY TO MANY FIELDS
        # ---------------------------------
        # Update the widgets for many to many fields
        for field in self._meta.model._meta.many_to_many:
            if field.name in self.fields:
                related_model = field.remote_field.model
                self.fields[field.name].widget = MultipleModelSelect(model=related_model)
    
        # ---------------------------------
        # FILE FIELDS
        # ---------------------------------
        # Update the widgets for the file fields
        self.file_fields = get_bloomerp_file_fields_for_model(self.model, output='list')

        # ---------------------------------
        # DATE AND DATETIME FIELDS
        # ---------------------------------
        for field_name, field in self.fields.items():
            if isinstance(field, forms.DateField):
                self.fields[field_name].widget = DateInput(attrs={'type': 'date'})
            elif isinstance(field, forms.DateTimeField):
                self.fields[field_name].widget = DateTimeInput(attrs={'type': 'datetime-local'})

        # ---------------------------------
        # JSON FIELDS
        # ---------------------------------
        # Update the widgets for the json fields
        for field_name, field in self.fields.items():
            # Check if the field is a JSONField
            model_field = self._meta.model._meta.get_field(field_name)
            
            if isinstance(model_field, (JSONField, DefaultJSONField)):
                # Apply the AceEditorWidget for JSON fields
                self.fields[field_name].widget = AceEditorWidget(language='json')

        # ---------------------------------
        # Remove created_by and updated_by fields
        # ---------------------------------
        if 'created_by' in self.fields:
            del self.fields['created_by']
        if 'updated_by' in self.fields:
            del self.fields['updated_by']
                
    def save(self, commit=True):
        '''
        Saves the form instance to the database

        '''

        instance = super(BloomerpModelForm, self).save(commit=False)

        # Check if the instance is new by checking if it has no primary key
        is_new_instance = instance.pk is None

        # Mark all temporary files as finalized after successful save
        def save_file_fields():
            if not instance.pk:
                raise ValueError("Instance must be saved before saving file fields")

            for field in self.file_fields:
                file:File = self.cleaned_data.get(field, None)
                if file:
                    # There is a new file so it has to be updated
                    file.persisted = True
                    file.content_type = ContentType.objects.get_for_model(self.model)
                    file.object_id = instance.pk
                    file.updated_by = self.user
                    file.created_by = self.user
                    file.save()
                else:
                    # There is no file, so we should delete the old file if it exists
                    old_file : File = getattr(instance, field, None)
                    if old_file:
                        old_file.delete()

        if commit:
            instance.save()
            save_file_fields()
        else:
            instance.save_file_fields = save_file_fields
            instance.is_new_instance = is_new_instance
        
        return instance

    def add_prefix(self, field_name):
        """
        Return the field name with a prefix appended.
        Overrides the default if the prefix contains "__".
        """
        if self.prefix and "__" in self.prefix:
            # Use "__" as the separator if the prefix contains "__"
            return f"{self.prefix}{field_name}"
        else:
            # Default behavior: use superclass method with hyphen separator
            return super().add_prefix(field_name)

# ---------------------------------
# Bloomerp Field Select Form
# ---------------------------------    
class BloomerpDownloadBulkUploadTemplateForm(forms.Form):
    file_type = forms.ChoiceField(choices=[('csv', 'CSV'), ('xlsx', 'Excel')])

    def __init__(self, model: Model = None, *args, **kwargs):
        super(BloomerpDownloadBulkUploadTemplateForm, self).__init__(*args, **kwargs)
        # Check if the content_type_id is provided in the request.POST
        content_type_id = kwargs.get('data', {}).get('content_type_id', None)
        if content_type_id:
            try:
                self.model = ContentType.objects.get(pk=content_type_id).model_class()
                print("Model:", self.model)
            except ContentType.DoesNotExist:
                self.model = None
                print("ContentType with id", content_type_id, "does not exist.")
        elif model:
            self.model = model
        else:
            raise ValueError("Model or content_type_id must be provided")

        if not self.model:
            raise ValueError("Model could not be determined.")

        # Init skip fields
        skip_fields = [
            'created_by',
            'updated_by',
            'datetime_created',
            'datetime_updated',
        ]

        # Model fields
        self.model_fields = []


        # Create model form and copy the fields
        ModelForm = modelform_factory(self.model, exclude=skip_fields)
        form_instance = ModelForm()

        # Add the fields to the form
        for field_name, field in form_instance.fields.items():
            self.fields[field_name] = forms.BooleanField(
                label=field.label,
                required=field.required,
                initial=field.required,
                disabled=field.required,
                help_text=field.help_text
            )

            # Add the field to the model fields
            self.model_fields.append(field_name)

        # Add content type id as hidden field
        self.fields['content_type_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=ContentType.objects.get_for_model(self.model).pk)

    def get_selected_fields(self):
        # Run the clean method to get the selected fields
        if self.is_valid():
            # Only include fields that are checked and in the model fields
            return [field for field in self.model_fields if self.cleaned_data.get(field, False)]
        return []
    

    

# ---------------------------------
# Bloomerp Bloomerp Model Form
# ---------------------------------
from bloomerp.models import File
class BloomerpModelFormDepr(forms.ModelForm):
    model:Model = None
    instance = None
    help_text_template = 'snippets/help_text_snippet.html'
    user:User = None
    form_name:str = None
    exception_models = ['file','contenttype']

    def __init__(self, user:User=None, *args, **kwargs):
        super(BloomerpModelFormDepr, self).__init__(*args, **kwargs)

        # Get instance from the form
        instance = kwargs.get('instance')

        # Set the form name 
        self.form_name = self._meta.model.__name__

        # Get the prefix from the form
        prefix = kwargs.get('prefix', None)

        # Set the instance to the form instance
        if instance:
            self.instance = instance


        if user and hasattr(user, 'get_content_types_for_user'):
            # Set the user to the form instance
            self.user = user

            # Get all of the different content types that the user has access to
            update_content_types = user.get_content_types_for_user(['change'])
            create_content_types = user.get_content_types_for_user(['add'])
        else:
            # This is a temporary solution to avoid errors when the user is not provided
            update_content_types = ContentType.objects.none()
            create_content_types = ContentType.objects.none()
            

        # Set the model instance to the form instance
        self.model = self._meta.model

        # Get the content type of the model
        content_type = ContentType.objects.get_for_model(self.model)
        file_content_type = ContentType.objects.get_for_model(File)

        # Get the application fields for the model
        self.application_fields = ApplicationField.objects.filter(content_type=content_type)

        # Get the file fields for the model
        file_fields_dict = get_file_fields_dict_for_model(self.model)

        self.file_fields = [field['name'] for field in file_fields_dict]

        # Add the file fields to the form
        for file_field in file_fields_dict:
            allowed_extensions = ', '.join(file_field['allowed_extensions'])
            attrs = {'class': 'form-control', 'accept': allowed_extensions}

            if instance:
                file_instance : File = getattr(instance, file_field['name'], None)
                if file_instance:
                    # Add the file field to the form
                    
                    self.fields[file_field['name']] = forms.FileField(required=file_field['required'], initial=file_instance.file, widget=forms.FileInput(attrs=attrs))
                    self.fields[file_field['name']].help_text = f'Current file: <a href="{file_instance.path}">{file_instance.name}</a> </br>Allowed extensions: {allowed_extensions}'
                else:
                    self.fields[file_field['name']] = forms.FileField(required=file_field['required'], widget=forms.FileInput(attrs=attrs))
                    self.fields[file_field['name']].help_text = f'Allowed extensions: {allowed_extensions}'
            else:
                self.fields[file_field['name']] = forms.FileField(required=file_field['required'], widget=forms.FileInput(attrs=attrs))
                self.fields[file_field['name']].help_text = f'Allowed extensions: {allowed_extensions}'
        
        # Get exception content types
        exception_content_types = self.get_exception_content_types().values_list('pk', flat=True)

        # Get foreign key fields for the model
        self.foreign_key_fields = self.application_fields.filter(field_type='ForeignKey').exclude(meta__related_model__in=exception_content_types).exclude(field='last_updated_by')

        # Add the foreign key fields to the form
        for foreign_key_field in self.foreign_key_fields:
            # Get the field name
            field = foreign_key_field.field
            
            if field not in self.fields.keys():
                continue

            # Set permissions
            create, update = False, False

            # Get the related model id, model and form
            related_model_id = foreign_key_field.meta['related_model']
            related_content_type = ContentType.objects.get(pk=related_model_id)
            ForeignModel = related_content_type.model_class()
            ForeignModelForm = modelform_factory(ForeignModel, exclude=['last_updated_by','user','avatar'], form=forms.ModelForm)            
            
            if related_content_type in create_content_types:
                create = True
                # Add create new item form
                create_form = ForeignModelForm(prefix=f'create_{field}')

                # Update fields
                updated_fields = {f'create_{field}_' + key: value for key, value in create_form.fields.items()}

                # Add the form to the form fields
                self.fields.update(updated_fields)

                for key, value in updated_fields.items():
                    self.fields[key].required = False
                    self.fields[key].widget.attrs['style'] = 'display: none;'
                    self.fields[key].label = ''
                    self.fields[key].help_text = ''

            # Check if user has update permission
            if related_content_type in update_content_types and instance and getattr(instance, field, None):
                update = True

                # Add update form
                update_form = ForeignModelForm(instance=getattr(instance, field), prefix=f'update_{field}')
                
                # Add update_field_id to the form
                update_form.fields[f'{field}_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=getattr(instance, field).pk)


                # Update fields & initial
                update_form.fields = {f'update_{field}_' + key: value for key, value in update_form.fields.items()}
                update_form.initial = {f'update_{field}_' + key: value for key, value in update_form.initial.items()}

                

                # Add the form to the form fields
                self.fields.update(update_form.fields)
                self.initial.update(update_form.initial)
                for key, value in update_form.fields.items():
                    self.fields[key].required = False
                    self.fields[key].widget.attrs['style'] = 'display: none;'
                    self.fields[key].label = ''
                    self.fields[key].help_text = ''
            
            
            # Add help text to the field
            self.fields[field].help_text = self.get_help_text(field, create, update, prefix)
                
        # Remove last_updated_by field
        if 'last_updated_by' in self.fields:
            del self.fields['last_updated_by']

    def get_exception_content_types(self):
        '''
        Returns the content types that should not be included in the foreign key fields
        '''
        return ContentType.objects.filter(model__in=self.exception_models)

    def get_content_type(self):
        '''
        Returns the content type of the model
        '''
        return ContentType.objects.get_for_model(self.model)

    def get_user(self):
        '''
        Returns the user that the form is associated with
        '''
        return self.user

    def get_help_text(self, field, create=False, update=False, prefix=None):
        '''
        Generates help text for the foreign key field
        '''
        if prefix is not None:
            prefix = prefix + '-'
        else:
            prefix = ''

        context = {
            'create' : create,
            'update' : update,
            'field' : field,
            'prefix' : prefix
        }

        template = get_template(self.help_text_template)
        help_text = template.render(context)
        
        return help_text

    def clean(self):
        cleaned_data = super().clean()

        # Handle file fields
        for field in self.file_fields:
            file = cleaned_data.get(field, None)
            
            if file and type(file) != UUID:
                # In this case the file is a new file
                try:
                    file = DjangoFile(file)
                    file_instance = self.clean_file_field(file, field)
                    cleaned_data[field] = file_instance
                except ValidationError as e:
                    self.add_error(field, e)
            elif file and type(file) == UUID:
                try:
                    cleaned_data[field] = File.objects.get(pk=file)
                except File.DoesNotExist:
                    self.add_error(field, ValidationError("File does not exist"))
                
        # Handle foreign key fields
        for foreign_key_field in self.foreign_key_fields:
            field_name = foreign_key_field.field

            create, update = False, False

            # Check if create / update checkbox is checked
            if self.data.get(f'create_{field_name}_checkbox', False) == 'on':
                create = True
            
            if self.data.get(f'update_{field_name}_checkbox', False) == 'on':
                update = True
            
            
            if create and update:
                raise ValidationError(f"Cannot create and update {field_name} at the same time")

            create_prefix = f'create_{field_name}_'
            update_prefix = f'update_{field_name}_'
            create_data = {key[len(create_prefix):]: val for key, val in cleaned_data.items() if key.startswith(create_prefix) and val}
            update_data = {key[len(update_prefix):]: val for key, val in cleaned_data.items() if key.startswith(update_prefix) and val}

            # Create a new instance only if data is provided
            if create_data and any(create_data.values()) and create:
                try:
                    cleaned_data[field_name] = self.clean_create_foreign_key(field_name, create_data)
                except ValidationError as e:
                    self.add_error(field_name, e)

            # Update existing instance only if changes are made
            elif update_data and any(update_data.values()) and update:
                try:
                    cleaned_data[field_name] = self.clean_update_foreign_key(field_name, update_data)
                except ValidationError as e:
                    self.add_error(field_name, e)
            # If neither create nor update is checked, the field should just be the selected instance
            # No need to do anything here

        return cleaned_data

    def clean_file_field(self, file:DjangoFile, file_field:str):
        # First check if the file is already associated with the instance
        if self.instance and getattr(self.instance, file_field, None):
            file_instance:File = getattr(self.instance, file_field)
            file_instance.file = file
            file_instance.name = file.name
            file_instance.save()
            
        else:
            if self.instance is None: 
                object_id = None
            else:
                object_id = self.instance.pk
            
            # Check if the file already exists
            # Get the datetime and makse sure that the file hasn't been uploaded in the last 60 minutes
            # This is to prevent the same file from being uploaded multiple times, which can happen when using django-formtools
            now = timezone.now()
            time_threshold = now - timedelta(minutes=60)

            file_instance = File.objects.filter(name=file.name,
                                                created_by=self.user,
                                                datetime_created__gte=time_threshold,
                                                object_id=object_id,
                                                ).first()
            if file_instance:
                return file_instance
            else:
                file_instance = File.objects.create(
                    created_by=self.user,
                    file=file,
                    name=file.name,
                    content_type=self.get_content_type(),
                    object_id=object_id
                )
        
        return file_instance

    def clean_create_foreign_key(self, field_name, data):
        ForeignModel:Model = self.fields[field_name].queryset.model
        ForeignModelForm = modelform_factory(ForeignModel, exclude=['last_updated_by', 'user'])
        form = ForeignModelForm(data=data,files=data)

        if form.is_valid():
            data = form.data
            # Prevent model from saving multiple times
            qs = ForeignModel.objects.filter(**data)
            if qs.exists():
                return qs.first()
            else:
                return form.save()

        else:
            raise ValidationError(f"Error creating new {ForeignModel.__name__}: {form.errors}")

    def clean_update_foreign_key(self, field_name, data):
        instance_id = data.get(f'{field_name}_id')

        ForeignModel = self.fields[field_name].queryset.model
        instance = ForeignModel.objects.get(pk=instance_id)
        ForeignModelForm = modelform_factory(ForeignModel, exclude=['last_updated_by', 'user', 'avatar'])
        form = ForeignModelForm(data, instance=instance)
        if form.is_valid():
            return form.save()
        else:
            raise ValidationError(f"Error updating {ForeignModel.__name__}: {form.errors}")
                 
    def save(self, commit=True):
        # Call the base class save method
        instance = super(BloomerpModelFormDepr, self).save(commit=commit)

        # Check if the instance is new by checking if it has no primary key
        is_new_instance = instance.pk is None

        # Assign the instance to self.instance for further processing
        self.instance = instance

        def update_file_fields():
            for field in self.file_fields:
                file = self.cleaned_data.get(field, None)
                if file:
                    file_instance = File.objects.get(pk=file.pk)
                    file_instance.content_type = self.get_content_type()
                    file_instance.object_id = self.instance.pk
                    file_instance.save()

        if is_new_instance and commit:
            update_file_fields()
        
        # If the instance is new and commit is False, we need to update the file fields after the instance has been saved
        elif is_new_instance and not commit:
            self.update_file_fields = update_file_fields
            instance.update_file_fields = update_file_fields
            instance.is_new_instance = is_new_instance
        
        return instance
    

# ---------------------------------
# Links select form
# ---------------------------------
class DetailLinksSelectForm(forms.Form):
    def __init__(self, content_type:ContentType, user:User, *args, **kwargs):
        super(DetailLinksSelectForm, self).__init__(*args, **kwargs)
        
        # Get all of the links that are available for the content type
        qs = Link.objects.filter(content_type=content_type, level='DETAIL') 
        
        for link in qs:
            if link.number_of_args() > 1:
                # Exclude links that require more than one argument
                qs = qs.exclude(pk=link.pk)
        
        # Get the links that the user has access to
        detail_view_links = UserDetailViewTab.get_detail_view_tabs(content_type=content_type, user=user).values_list('link_id', flat=True)

        self.fields['links'] = forms.ModelMultipleChoiceField(
            queryset=qs,
            widget=forms.CheckboxSelectMultiple
        )

        self.fields['links'].initial = detail_view_links

        # Add content type id as hidden field
        self.fields['content_type_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=content_type.pk)

        self.user = user


    def save(self) -> None:
        content_type_id = self.cleaned_data.get('content_type_id')
        content_type = ContentType.objects.get(pk=content_type_id)
        links = self.cleaned_data.get('links')

        # Get existing UserDetailViewTab objects for the user and content type
        existing_tabs = UserDetailViewTab.objects.filter(user=self.user, link__content_type=content_type)

        # Determine which links need to be added and which need to be removed
        existing_link_ids = set(existing_tabs.values_list('link_id', flat=True))
        selected_link_ids = set(links.values_list('id', flat=True))

        # Links to add
        links_to_add = selected_link_ids - existing_link_ids
        # Links to remove
        links_to_remove = existing_link_ids - selected_link_ids

        # Add new UserDetailViewTab objects
        for link_id in links_to_add:
            UserDetailViewTab.objects.create(user=self.user, link_id=link_id)

        # Remove UserDetailViewTab objects
        UserDetailViewTab.objects.filter(user=self.user, link_id__in=links_to_remove).delete()

        print("Links saved")


class ListViewFieldsSelectForm(forms.Form):
    def __init__(self, content_type:ContentType, user:User, *args, **kwargs):
        super(ListViewFieldsSelectForm, self).__init__(*args, **kwargs)
        
        # Get all of the Application fields that are available for the content type
        qs = ApplicationField.objects.filter(content_type=content_type)
        
        
        # Get the links that the user has access to
        application_fields = UserListViewPreference.objects.filter(user=user, application_field__in=qs).values_list('application_field_id', flat=True)
        
        self.fields['fields'] = forms.ModelMultipleChoiceField(
            queryset=qs,
            widget=forms.CheckboxSelectMultiple
        )

        self.fields['fields'].initial = application_fields

        # Add content type id as hidden field
        self.fields['content_type_id'] = forms.IntegerField(widget=forms.HiddenInput(), initial=content_type.pk)

        self.user = user


    def save(self) -> None:
        content_type_id = self.cleaned_data.get('content_type_id')
        content_type = ContentType.objects.get(pk=content_type_id)
        fields = self.cleaned_data.get('fields')

        # Get existing UserListViewPreference objects for the user and content type
        existing_preferences = UserListViewPreference.objects.filter(user=self.user, application_field__content_type=content_type)

        # Determine which fields need to be added and which need to be removed
        existing_field_ids = set(existing_preferences.values_list('application_field_id', flat=True))
        selected_field_ids = set(fields.values_list('id', flat=True))

        # Fields to add
        fields_to_add = selected_field_ids - existing_field_ids
        # Fields to remove
        fields_to_remove = existing_field_ids - selected_field_ids

        # Add new UserListViewPreference objects
        for field_id in fields_to_add:
            UserListViewPreference.objects.create(user=self.user, application_field_id=field_id)

        # Remove UserListViewPreference objects
        UserListViewPreference.objects.filter(user=self.user, application_field_id__in=fields_to_remove).delete()




from django.db import models
from django.core.exceptions import ValidationError
from django.urls import reverse
from bloomerp.models.core import BloomerpModel
from django.contrib.contenttypes.models import ContentType


# ---------------------------------
# SQL Query Filter Model
# ---------------------------------
from bloomerp.utils.sql import SqlQueryExecutor
class SqlQueryFilter(BloomerpModel):
    '''
    A model to store filters for KPIs. Ech filter has a name and a query.
    '''
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_sql_query_filter'
        verbose_name = 'SQL Query Filter'
        verbose_name_plural = 'SQL Query Filters'

    FILTER_TYPE_CHOICES = [
        ('absolute', 'Absolute'),
        ('integer', 'Integer'),
        ('float', 'Float'),
        ('date', 'Date'),
        ('boolean', 'Boolean'),
        ('string', 'String'),
        ('select', 'Select'),
        ('model_select', 'Model Select')
    ]

    name = models.CharField(max_length=255)
    query = models.TextField()
    filter_type = models.CharField(max_length=255, default='absolute', choices=FILTER_TYPE_CHOICES)
    filter_options = models.JSONField(blank=True, null=True)

    # Query executor
    executor = SqlQueryExecutor()

    def __str__(self):
        return self.name
    
    def clean(self):
        '''
        Ensure that the filter query is lowercase
        '''
        self.query = self.query.lower()

        # Check if the query is safe
        if not self.executor.is_safe(self.query):
            raise ValidationError('Unsafe query')
        
        if self.filter_type == 'absolute':
            # No need to check the query
            pass
        elif self.filter_type == 'integer':
            # Check if a default value is provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (default)')

            if 'default' not in self.filter_options:
                raise ValidationError('Default value is required for integer filters')
            else:
                # Check if the default value is an integer
                try:
                    int(self.filter_options['default'])
                except ValueError:
                    raise ValidationError('Default value must be an integer')
        elif self.filter_type == 'float':
            # Check if a default value is provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (default)')

            if 'default' not in self.filter_options:
                raise ValidationError('Default value is required for float filters')
            else:
                # Check if the default value is a float
                try:
                    float(self.filter_options['default'])
                except ValueError:
                    raise ValidationError('Default value must be a float')
        elif self.filter_type == 'date':
            # Check if a default value is provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (default)')


            if 'default' not in self.filter_options:
                raise ValidationError('Default value is required for date filters')
            else:
                # Check if the default value is a date
                try:
                    from datetime import datetime
                    datetime.strptime(self.filter_options['default'], '%Y-%m-%d')
                except ValueError:
                    raise ValidationError('Default value must be a date in the format YYYY-MM-DD')
        elif self.filter_type == 'boolean':
            # Check if a default value is provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (default)')

            if 'default' not in self.filter_options:
                raise ValidationError('Default value is required for boolean filters')
            else:
                # Check if the default value is a boolean
                if self.filter_options['default'] not in [0, 1]:
                    raise ValidationError('Default value must be either 0 (False) or 1 (True)')
        elif self.filter_type == 'string':
            # Check if a default value is provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (default)')

            if 'default' not in self.filter_options:
                raise ValidationError('Default value is required for string filters')
            else:
                # Check if the default value is a string
                if not isinstance(self.filter_options['default'], str):
                    raise ValidationError('Default value must be a string')
        elif self.filter_type == 'select':
            # Check if options are provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (options)')

            if 'options' not in self.filter_options:
                raise ValidationError('Options are required for select filters')
            else:
                # Check if the options are a list
                if not isinstance(self.filter_options['options'], list):
                    raise ValidationError('Options must be a list')
                self.filter_options['default'] = self.filter_options['options'][0]    
            
        elif self.filter_type == 'model_select':
            # Check if options are provided
            if not self.filter_options:
                raise ValidationError('Options are required for integer filters (content_type_id)')

            if 'content_type_id' not in self.filter_options:
                raise ValidationError('Content type ID is required for model select filters')
            else:
                # Check if the model exists in the content types
                if not ContentType.objects.filter(pk=self.filter_options['content_type_id']).exists():
                    raise ValidationError('Invalid content type ID')
                # Set default value 
                try:
                    self.filter_options['default'] = ContentType.objects.get(pk=self.filter_options['content_type_id']).model_class().objects.first().pk
                except:
                    self.filter_options['default'] = None

        else:
            raise ValidationError('Invalid filter type')
        
    def format_query_filter(self, value=None) -> str:
        '''
        Formats the query filter with the provided value, if no value is provided, the default value is used.
        '''
        if self.filter_type == 'absolute':
            return self.query
        else:
            formatted_query = self.query.format(value=value if value is not None else self.filter_options['default'])

        return formatted_query
    
    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)  

# ---------------------------------
# SQL Query Model
# ---------------------------------
from django.contrib.auth.hashers import make_password
from bloomerp.models.fields import CodeField
class SqlQuery(BloomerpModel):
    '''
    A model to store SQL queries that can be used in Widgets
    '''
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'sql_query'
        verbose_name = 'SQL Query'
        verbose_name_plural = 'SQL Queries'
        permissions = [
            ('execute_sql_query', 'Can execute SQL queries')
            # Maybe add more permissions here corresponding to the actions that can be performed on the query
        ]

    name = models.CharField(max_length=255)
    query = CodeField(language='sql', help_text='SQL Query to execute')
    filters = models.ManyToManyField(SqlQueryFilter, blank=True)

    # String fields
    search_fields = ['name']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize the query executor
        if self.name:
            self.executor = SqlQueryExecutor(
                cache_time=60,
                cache_id=make_password(self.name, 'md5')
            )
        else:
            self.executor = SqlQueryExecutor()

    @property
    def is_safe(self):
        '''
        Check if the query is a SELECT statement
        '''
        return self.executor.is_safe(self.query)

    @property
    def result_dataframe(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_to_df(self.query)
    
    @property
    def result_dict(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_to_dict(self.query)
    
    @property
    def result_raw(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_raw(self.query)
    
    @property
    def result_value(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_to_first_value(self.query)


    def clean(self) -> None:
        '''
        Clean method does the following:
        - Check if the query is safe
        - Check if the query actually returns a result
        '''
        # Check if the query is safe
        # In the future, we can allow unsafe queries for more advanced users
        if not self.is_safe:
            raise ValidationError('Unsafe query')
        
        # Check if the query returns a result
        result = self.executor.is_valid(self.query)
        if result != True:
            raise ValidationError(result)

    def __str__(self):
        return self.name

    def apply_filters(self, filters: list):
        '''
        Apply filters to the query
        '''
        filter_qs = SqlQueryFilter.objects.filter(pk__in=filters)

        # Remove ; from the query
        self.query = self.query.replace(';', '')

        # Lowercase the query
        self.query = self.query.lower()

        # Generate the filter query string
        filter_query = ' and '.join([filter for filter in filters])

        # Check if query has where clause
        if 'where' in self.query:
            # Add the filter_query after an AND if WHERE clause exists
            self.query = self.query.replace('where', f'where {filter_query} and')
        else:
            # Add WHERE clause if it does not exist
            self.query += f' where {filter_query}'

        # Append the semicolon to the end of the query
        self.query += ';'

        return self.query



# ---------------------------------
# KPI Model
# ---------------------------------
from pandas import DataFrame
class Widget(BloomerpModel):
    OUTPUT_TYPE_CHOICES = [
        ('table','Table'),
        ('value','Value'),
        ('line','Line chart'),
        ('bar','Bar chart'),
        ('pie','Pie chart'),
        ('scatter','Scatter chart'),
        ('histogram','Histogram')
    ]

    VALUE_AGGREGATE_CHOICES = [
        ('sum', 'Sum'),
        ('average', 'Average'),
        ('count', 'Count'),
        ('min', 'Minimum'),
        ('max', 'Maximum'),
        ('first', 'First value')
    ]

    name = models.CharField(max_length=255) # Name of the KPI
    query = models.ForeignKey(SqlQuery, on_delete=models.CASCADE) # Query to execute
    description = models.TextField(blank=True,null=True) # Description of the KPI
    output_type = models.CharField(choices=OUTPUT_TYPE_CHOICES, max_length=10)
    options = models.JSONField(blank=True,null=True)


    string_search_fields = ['name', 'description']

    def __str__(self):
        return self.name
    
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_widget'
    
    def clean(self):
        # SAFETY CHECKS
        if not self.query.is_safe:
            raise ValidationError('Unsafe query')

        # Get columns
        dataframe = self.query.result_dataframe
        columns = dataframe.columns.tolist()

        # TABLE SPECIFIC VALIDATION
        if self.output_type == 'table':
            if self.options is None:
                raise ValidationError('Options are required for table KPIs')

            if 'columns' not in self.options:
                raise ValidationError('Columns are required for table KPIs')
            
            for col in self.options['columns']:
                if col not in columns:
                    raise ValidationError(f'Column {col} does not exist in the query result')
            if 'limit' not in self.options:
                self.options['limit'] = 10

            

        # GRAPH SPECIFIC VALIDATION
        if (self.output_type == 'bar' or self.output_type == 'line' or self.output_type == 'scatter') and self.options is None:
            raise ValidationError('Options are required for chart KPIs')

        
        # BAR / LINE / SCATTER SPECIFIC VALIDATION
        if self.output_type == 'bar' or self.output_type == 'line' or self.output_type == 'scatter':
            if 'x' not in self.options:
                raise ValidationError('X-axis is required for bar chart KPIs')
            else:
                if self.options['x'] not in columns:
                    raise ValidationError(f'Column {self.options["x"]} does not exist in the query result')
            
            if 'y' not in self.options:
                raise ValidationError('Y-axis is required for bar chart KPIs')
            else:
                if self.options['y'] not in columns:
                    raise ValidationError(f'Column {self.options["y"]} does not exist in the query result')

            if 'group_by' in self.options:
                if self.options['group_by'] not in columns:
                    raise ValidationError(f'Column {self.options["group_by"]} does not exist in the query result')

        # PIE SPECIFIC VALIDATION     
        if self.output_type == 'pie':
            if 'x' not in self.options:
                raise ValidationError('X-axis is required for pie chart KPIs')
            else:
                if self.options['x'] not in columns:
                    raise ValidationError(f'Column {self.options["x"]} does not exist in the query result')
            
            if 'group_by' not in self.options:
                raise ValidationError('Groupby column is required for pie chart KPIs')
            else:
                if self.options['group_by'] not in columns:
                    raise ValidationError(f'Column {self.options["y"]} does not exist in the query result')
                
        # HISTOGRAM SPECIFIC VALIDATION
        if self.output_type == 'histogram':
            if 'x' not in self.options:
                raise ValidationError('X-axis is required for histogram KPIs')
            else:
                if self.options['x'] not in columns:
                    raise ValidationError(f'Column {self.options["x"]} does not exist in the query result')
                
        # VALUE SPECIFIC VALIDATION
        if self.output_type == 'value':
            if 'color' not in self.options:
                self.options['color'] = 'gray'
            
            if 'prefix' not in self.options:
                self.options['prefix'] = ''

            if 'suffix' not in self.options:
                self.options['suffix'] = ''
            
            if 'icon' not in self.options:
                self.options['icon'] = 'bi bi-bar-chart'

            if 'column' not in self.options:
                raise ValidationError('Column is required for value KPIs')

            if 'aggregate' not in self.options:
                self.options['aggregate'] = 'first'
            
            if self.options['aggregate'] not in [agg[0] for agg in self.VALUE_AGGREGATE_CHOICES]:
                raise ValidationError('Invalid aggregate type')
            
    @property
    def result_value(self):
        '''
        Execute the query and return the result
        '''
        dataframe:DataFrame = self.query.result_dataframe

        if self.output_type == 'value':
            column = self.options['column']
            aggregate = self.options['aggregate']

            try:
                if aggregate == 'sum':
                    return dataframe[column].sum()
                elif aggregate == 'average':
                    return dataframe[column].mean()
                elif aggregate == 'count':
                    return dataframe[column].count()
                elif aggregate == 'min':
                    return dataframe[column].min()
                elif aggregate == 'max':
                    return dataframe[column].max()
                elif aggregate == 'first':
                    return dataframe[column].iloc[0]
            except:
                self.options['aggregate'] = 'first'
                self.save()
                return dataframe[column].iloc[0]
        else:
            return 'Value result not available for this KPI type'

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)




from django import forms
from bloomerp.models import DocumentTemplate
from django.apps import apps

# ---------------------------------
# Free Variable Form
# ---------------------------------
class FreeVariableForm(forms.Form):
    def __init__(self, document_template: DocumentTemplate, *args, **kwargs):
        super().__init__(*args, **kwargs)

        free_variables = document_template.free_variables.all()
        for variable in free_variables:
            field_kwargs = {
                'label': variable.name,
                'required': variable.required,
            }
            if variable.variable_type == 'date':
                field_kwargs['widget'] = forms.DateInput(
                    attrs={'type': 'date'})
                field = forms.DateField(**field_kwargs)
            elif variable.variable_type == 'boolean':
                field = forms.BooleanField(**field_kwargs)

            elif variable.variable_type == 'integer':
                field = forms.IntegerField(**field_kwargs)

            elif variable.variable_type == 'float':
                field = forms.FloatField(**field_kwargs)

            elif variable.variable_type == 'text':
                field = forms.CharField(**field_kwargs)
            elif variable.variable_type == 'list':
                # Create options from choices
                options = variable.options.split(',')
                field_kwargs['choices'] = [(option, option)
                                           for option in options]

                field = forms.ChoiceField(**field_kwargs)
            elif variable.variable_type == 'model':
                # Retrieve the model
                try:
                    model = variable.options.split(',')
                    # Get the model class using the apps module without specifying app_label
                    model_class = apps.get_model(
                        app_label=model[0], model_name=model[1])

                    # Get a queryset for the model
                    queryset = model_class.objects.all()

                    # create the field
                    field = forms.ModelChoiceField(queryset=queryset)
                except:
                    # Handle the case where the model is not found
                    pass

            else:
                field = forms.CharField(**field_kwargs)

            self.fields[variable.slug] = field




from django.apps import apps
from django.shortcuts import render
from django.http import HttpRequest, HttpResponse
from bloomerp.models import BloomerpModel
from bloomerp.utils.router import route
from bloomerp.models import Link
from bloomerp.utils.models import search_models_by_query

@route('search_results')
def search_results(request: HttpRequest) -> HttpResponse:
    query = request.GET.get('search_results_query', '')  # Retrieve the search query from the GET parameters
    results = []

    links = None

    if query == '':
        return HttpResponse('<li class="dropdown-item">No query provided</li>')

    if query.startswith('/') and query[1] != '/':
        # In this case we want to return links for a specific model
        # Name is optional and comes after a : in the query
        sub_query = query[1:]

        name = None
        if ':' in sub_query:
            sub_query, name = sub_query.split(':')

        content_types = search_models_by_query(sub_query)
        list_level_links = []

        for content_type in content_types:
            links = Link.objects.filter(content_type=content_type, level='LIST')
            if name:
                links = links.filter(name__icontains=name)
            
            list_level_links.append({
                'model_name': content_type.model_class()._meta.verbose_name,
                'links': links
            })
        return render(request, 'components/search_results.html', {'list_level_links': list_level_links})
    
    if query.startswith('//'):
        # In this case, we want to return app level links
        sub_query = query[2:]
        app_level_links = Link.objects.filter(level='APP', name__icontains=sub_query)

        return render(request, 'components/search_results.html', {'app_level_links': app_level_links})

    # Get all models from the app
    all_models = apps.get_models()
    # Iterate over all models
    for model in all_models:
        # Check if user has permission to view the model
        if not request.user.has_perm(f'{model._meta.app_label}.view_{model._meta.model_name}'):
            continue

        # Check if the model inherits from BloomerpModel and allows string search
        if issubclass(model, BloomerpModel) and getattr(model, 'allow_string_search', False):
            # Perform string search using the static method
            matching_objects = model.string_search(query)

            # If there are matching objects, add them to the results
            if matching_objects.exists():
                if len(matching_objects) > 5:
                    matching_objects = matching_objects[0:5]

                results.append({
                    "model_name": model._meta.verbose_name_plural,
                    "objects": matching_objects
                })


    context = {
        "results": results,
        "query": query
    }

    return render(request, 'components/search_results.html', context)



from django import template
from django.db.models.manager import Manager
from django.db.models import Model
from bloomerp.utils.models import model_name_plural_underline, get_model_dashboard_view_url, get_list_view_url
from django.urls import reverse 
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import Link, Widget
from django.utils.safestring import mark_safe
import uuid
from bloomerp.models import Bookmark, User

register = template.Library()

@register.filter(name='get_field_value')
def get_field_value(obj:Model, field_name:str):
    """
    Custom template filter to get the value of a field by its name.
    
    Example usage:
    {{ object|get_field_value:field_name }}

    """
    if '.' in field_name:
        attributes = field_name.split('.')
        value = obj
        for attribute in attributes:
            try:
                value = getattr(value, attribute)
            except Exception as e:
                return e
        return value
    try:
        return getattr(obj, field_name)
    except Exception as e:
        return e

from django.contrib.contenttypes.models import ContentType
@register.filter(name='get_foreign_key_value')
def get_foreign_key_value(obj:Model,field_name):
    '''
    
    '''

    try:
        resp = ''

        value = getattr(obj, field_name)
        if isinstance(value, Manager):
            qs = value.all()

            if not qs:
                return 'None'

            content_type_id = ContentType.objects.get_for_model(qs.model).pk
      

            resp += f'<a href="/simple-list-view/?content_type_id={content_type_id}&field_name={obj._meta.model_name}&field_value={obj.pk}">'
            for item in qs[:2]:
                resp+= item.__str__() + ', '
            resp += '... </a>'
            return resp
        else:
            try:
                return f'<a href="{value.get_absolute_url()}">{value}</a>'
            except:
                return value
    except Exception as e:
        return e

@register.filter(name='get_dict_value')
def get_dict_value(dictionary:dict, key:str):
    '''
    Returns the value of a key in a dictionary.

    Example usage:
    {{ dictionary|get_dict_value:key }}
    '''

    return dictionary.get(key)

@register.filter
def model_name(obj:Model):
    '''
    Returns the model name of an object.

    Example usage:
    {{ object|model_name }}
    
    '''
    return obj._meta.model_name

@register.filter
def model_name_plural(obj:Model):
    '''
    Returns the model verbose name of an object.

    Example usage:
    {{ object|model_name_plural }}
    
    '''
    return obj._meta.verbose_name_plural

@register.filter
def model_dashboard_url(content_type:ContentType):
    '''
    Returns the model app dashboard URL of an object.

    Example usage:
    {{ object|model_app_dashboard_url }}
    
    '''
    return reverse(get_model_dashboard_view_url(content_type.model_class()))

@register.filter
def model_name_plural_from_content_type(content_type:ContentType):
    '''
    Returns the model verbose name of an object.

    Example usage:
    {{ object|model_name_plural }}
    
    '''
    return content_type.model_class()._meta.verbose_name_plural

@register.filter
def length(obj) -> int:
    '''
    Returns the length of an object.

    Example usage:
    {{ object|length }}
    
    '''
    return len(obj)

@register.filter
def percentage(value, arg):
    try:
        value = int(value) / int(arg)
        return value*100
    except (ValueError, ZeroDivisionError):
        return None
    
@register.filter
def get_link_by_id(id:int):
    '''
    Returns the link object with the given id.

    Example usage:
    {{ id|get_link }}

    or 

    {% with id|get_link as link %}
    {% if link %}
    {{ link.name }}
    {% endif %}
    {% endwith %}

    '''
    try:
        return Link.objects.get(pk=id)
    except:
        return None
    
@register.filter
def get_widget_by_id(id:int):
    '''
    Returns the widget object with the given id.

    Example usage:
    {{ id|get_widget }}

    or 

    {% with id|get_widget as widget %}
    {% if widget %}
    {{ widget.name }}
    {% endif %}
    {% endwith %}

    '''
    try:
        return Widget.objects.get(pk=id)
    except:
        return None
    

@register.inclusion_tag('snippets/breadcrumb.html')
def breadcrumb(title:str, model:Model):
    '''
    Returns a breadcrumb navigation.

    Example usage:
    {% breadcrumb links %}
    '''
    list_view_url = get_list_view_url(model)
    model_name_plural = model._meta.verbose_name_plural.title()

    return {
        'title' : title,
        'list_view_url' : list_view_url,
        'model_name_plural' : model_name_plural	
    }
    
@register.inclusion_tag('snippets/workspace_item.html')
def workspace_item(item:dict):
    '''
    Returns a workspace item.

    Example usage:
    {% workspace_item item %}
    '''
    # generate random id for each item
    item['id'] = uuid.uuid4()

    return {'item': item}


@register.simple_tag(name='render_link')
def render_link(link_id:int):
    '''
    Returns a link object.

    Example usage:
    {% render_link link_id %}
    '''
    try:
        link = Link.objects.get(pk=link_id)
        if link.is_external_url():
            return mark_safe(f'<a link-id="{link.pk}" class="pointer text-primary link-item" href="https://{link.url}" target="_blank">{link.name}</a>')
        elif link.is_absolute_url:
            return mark_safe(f'<a link-id="{link.pk}" class="pointer text-primary link-item" hx-get="{link.url}" hx-target="#main-content" hx-push-url="true">{link.name}</a>')
        elif not link.requires_args():
            return mark_safe(f'<a link-id="{link.pk}" class="pointer text-primary link-item" hx-get="{reverse(link.url)}" hx-target="#main-content" hx-push-url="true">{link.name}</a>')
        else:
            return mark_safe("<p>Link requires arguments</p>")
    except:
        return mark_safe("<p>Link not found</p>")

@register.inclusion_tag('components/bookmark.html')
def render_bookmark(object:Model, user:User, size:int, target:str):
    '''
    Returns a bookmark object.

    Example usage:
    {% render_bookmark object user size target %}
    '''
    # Get the content_type_id and object_id from the request
    content_type_id = ContentType.objects.get_for_model(object).pk
    
    # Check if the bookmark allready exists
    bookmarked = Bookmark.objects.filter(user=user, content_type_id=content_type_id, object_id=object.pk).exists()

    return {
        'bookmarked': bookmarked,
        'content_type_id': content_type_id,
        'object_id': object.pk,
        'target' : target,
        'size': size
    }



# Generated by Django 5.1.1 on 2024-09-18 17:46

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "bloomerp",
            "0015_rename_created_at_documenttemplate_datetime_created_and_more",
        ),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Comment",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                ("object_id", models.PositiveIntegerField()),
                ("comment", models.TextField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_comment",
                "managed": True,
            },
        ),
    ]



from django.shortcuts import render
from bloomerp.utils.router import BloomerpRouter
from .models import Employee
from bloomerp.views.mixins import HtmxMixin
from django.views import View

# Create your views here.

router = BloomerpRouter()

@router.bloomerp_route(
    path='send-emails',
    name='Send Emails',
    description='Send emails to employees',
    route_type='list',
    url_name='send_emails',
    models=Employee
)
class SendEmailsView(HtmxMixin, View):
    template_name = 'email_employees.html'
    model = Employee


    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        employee_list = Employee.objects.all()
        context['employee_list'] = employee_list
        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data()
        employee_list = Employee.objects.all()
        context['employee_list'] = employee_list
        return render(request, self.template_name, context)
        
    def get_context_data(self, **kwargs):
        return {}


from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.contrib.contenttypes.models import ContentType
from django.forms.models import modelform_factory
from bloomerp.forms.core import BloomerpModelForm
from bloomerp.models import User
from django.contrib.auth.decorators import login_required

@login_required
@route('object_model_form')
def object_model_form(request:HttpRequest) -> HttpResponse:
    '''
    Component to create an object of a model.
    Requires content_type_id and form_prefix in the query parameters.
    '''
    # Get the content type id and form prefix from the query parameters
    content_type_id = request.GET.get('content_type_id')
    form_prefix = request.GET.get('form_prefix','')
    object_id = request.GET.get('object_id', None)
    field_name = request.GET.get('field_name', None)

    # Add permission check here
    user : User = request.user


    if not content_type_id:
        return HttpResponse('content_type_id required in the query parameters', status=400)
    
    # Get the model and create a form
    model = ContentType.objects.get(id=content_type_id).model_class()

    # Permission check
    if not user.has_perm(f'{model._meta.app_label}.add_{model._meta.model_name}') and not user.has_perm(f'{model._meta.app_label}.change_{model._meta.model_name}'):
        return HttpResponse('User does not have permission to add objects of this model')

    if field_name:
        fields = [field_name]
        Form = modelform_factory(model, fields=fields, form=BloomerpModelForm)
    else:
        Form = modelform_factory(model, fields='__all__', form=BloomerpModelForm)

    if request.method == 'POST':
        form = Form(data=request.POST, files=request.FILES, prefix=form_prefix, model=model)
        if form.is_valid():
            form.save()
            return HttpResponse('Form submitted successfully')
    else:
        if object_id:
            instance = model.objects.get(id=object_id)
            form = Form(instance=instance, prefix=form_prefix, model=model, user=user)
        else:
            form = Form(prefix=form_prefix, model=model, user=user)

    return render(request, 'components/object_model_form.html', {'form': form})


from openai import OpenAI
from django.conf import settings


class BloomerpOpenAI:
    def __init__(self):
        API_KEY = settings.BLOOMERP_SETTINGS.get('BLOOMERP_OPENAI_API_KEY', None)

        self.client = OpenAI(api_key=API_KEY)
        self.model = "gpt-4o"
    
    def is_valid_key(self):
        '''Checks if the OpenAI API key is valid.'''
        try:
            self.client.models.list()
            return True
        except Exception:
            return False


    def create_sql_query(self,
                        query: str,
                        db_tables_and_columns: list[tuple[str, list[str]]]
                        ) -> str:
        '''
        This function creates a SQL query from a natural language query.
        '''
        system_content = '''
        You are a helpful assistant that helps me to create SQL queries from natural language.
        The output should be a SQL query using sqlite3 syntax, without any explanation. Dont include in the output ```sql ... ```, just the raw SQL query.
        Here are the database tables, columns and datatypes for each column in the database:
        '''

        for table in db_tables_and_columns:
            system_content += f'\n{table[0]}: '

            for column in table[1]:
                system_content += f'{column[0]} ({column[1]}), '
        

        response = self.client.chat.completions.create(
            model = self.model,
            messages = [
                {'role': 'system', 'content': system_content},
                {'role': 'user', 'content': query},
            ]
        )


        sql_query = response.choices[0].message.content

        # Remove ```sql ... ``` from the output
        sql_query = sql_query.replace('```sql\n', '').replace('```', '')

        return sql_query



from xhtml2pdf import pisa
import requests
import os
from io import BytesIO

def fetch_image(image_url):
    """Helper function to download image from a URL and save it temporarily."""
    try:
        response = requests.get(image_url)
        if response.status_code == 200:
            # Save the image temporarily in memory
            temp_image_path = os.path.join(os.getcwd(), "temp_image.jpg")
            with open(temp_image_path, 'wb') as img_file:
                img_file.write(response.content)
            return temp_image_path
        return None
    except Exception as e:
        print(f"Error fetching image: {e}")
        return None

def link_callback(uri, rel):
    """Link callback function to handle external images."""
    if uri.startswith("http://") or uri.startswith("https://"):
        # If the URI is an external image, fetch and return its local path
        return fetch_image(uri)
    else:
        # If it's a local file path, resolve it and return
        return uri

def generate_pdf(
        html_content:str, 
        css_content:str|None=None,
        font_url=None, 
        page_margin=1, 
        is_landscape=False
        )->bytes:
    '''Generate a PDF from HTML content with optional CSS styling and custom font.'''

    # Adjust the page margin and orientation via inline CSS
    orientation = "landscape" if is_landscape else "portrait"

    # Define the CSS for the PDF
    if not css_content:
        css_content = ""

    css = f"""
    <style>
        @page {{
            size: {orientation};
            margin: {page_margin}in;

            
        }}
        body {{
            font-family: 'CustomFont', sans-serif;
        }}

        {css_content}

    </style>
    """

    # Add Google Fonts embedding if a font URL is provided
    if font_url:
        css += f"<link href='{font_url}' rel='stylesheet'>"

    # Define the footer HTML
    footer_html = ""

    # Combine CSS with the HTML content
    full_html = css + html_content + footer_html

    # Create an in-memory buffer
    pdf_buffer = BytesIO()

    # Create the PDF using xhtml2pdf's pisa
    pisa_status = pisa.CreatePDF(
        src=full_html,
        dest=pdf_buffer,
        link_callback=link_callback  # For handling images from URLs
    )

    # Check if the PDF was generated successfully
    if pisa_status.err:
        print("Error occurred while creating PDF")
        return None
    else:
        print("PDF generated successfully")
        pdf_buffer.seek(0)  # Move the buffer cursor to the beginning
        return pdf_buffer.getvalue()


from typing import Any
from django.db.models.base import Model as Model
from django.db.models.query import QuerySet
from django.forms import formset_factory
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render, redirect
from django.urls import reverse, reverse_lazy
from django.contrib import messages
#from modules.shared_utils.mixins import ForeignRelationshipMixin
#from modules.shared_utils.shared_views import BaseListView
from bloomerp.forms.auth import UserDetailViewPreferenceForm, UserDetailViewPreferenceFormset
from bloomerp.forms.core import ListViewFieldsSelectForm
from django.contrib.contenttypes.models import ContentType
from django.core.paginator import Paginator
from bloomerp.views.mixins import HtmxMixin
from bloomerp.models import (
    UserDetailViewPreference,
    User
    )
from django.views.generic.edit import CreateView
from django.views.generic.detail import DetailView
#from .models import User, UserDetailViewPreference, UserListViewPreference
from bloomerp.utils.models import model_name_plural_underline
from django.contrib.auth.mixins import PermissionRequiredMixin
class UserDetailBaseView(PermissionRequiredMixin, DetailView):
    tabs = [
        #{'name': 'Overview', 'route': 'users_overview', 'id': False},
        #{'name': 'Change password', 'route': 'users_change_password', 'id': False},
        #{'name': 'Notification settings', 'route': 'notification_settings', 'id': False},
        #{'name': 'Object view settings', 'route': 'detail_view_preference', 'id': False},
        #{'name': 'List view settings', 'route': 'list_view_preference', 'id': False},
        #{'name': 'KPI Dashboard settings', 'route': 'widget_dashboard_view_preference', 'id': False},
        #{'name': 'KPI ListView settings', 'route': 'widget_list_view_preference', 'id': False},
        #{'name': 'Change details', 'route': 'users_settings', 'id': False},
    ]
    module_info = {'root': 'dashboard_homepage', 'name': 'Users'}
    model = User

    permission_required = []
    
    def get_object(self) -> Model:
        return self.request.user

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Set tabs
        context['tabs'] = self.tabs

        # Set model name
        context['model'] = 'user'

        # Set module info
        context['module'] = self.module_info
        self.context = context
        return context


class UserOverviewView(UserDetailBaseView):
    template_name = 'base_detail_overview_view.html'

    left_column = [
        {'field_name': 'Username','field_value':'username'},
        {'field_name': 'First name','field_value':'first_name'},
        {'field_name': 'Last name','field_value':'last_name'},
        {'field_name': 'Email','field_value':'email'},
    ]
    middle_column = [
        
    ]


    extra_context = {
        'left_column':left_column,
        'middle_column':middle_column,
    }

from django.contrib.auth.views import PasswordChangeView
from django.contrib.messages.views import SuccessMessageMixin
class UserPasswordChangeView(
        SuccessMessageMixin, 
        UserDetailBaseView, 
        PasswordChangeView
    ):
    template_name = 'base_detail_form_view.html'
    success_url = reverse_lazy("users_overview")
    success_message = 'Password updated succesfully.'

from django.views.generic.edit import UpdateView
class UserSettingsView(SuccessMessageMixin, UserDetailBaseView, UpdateView):
    template_name = 'base_detail_form_view.html'
    success_url = reverse_lazy("users_overview")
    success_message = 'Settings updated succesfully.'
    fields = ['first_name','last_name','email','avatar']

class ForeignRelationshipMixin:
    pass

class BaseListView:
    pass

class Notification:
    pass

class UserModelNotificationSetting:
    pass



# -------------------------------------------
# NEW SHIT
from bloomerp.utils.router import BloomerpRouter
from bloomerp.views.core import BloomerpBaseDetailView
from bloomerp.views.mixins import BloomerpModelFormViewMixin
from django.contrib.auth.forms import PasswordChangeForm
from django.views.generic.edit import FormView

class ProfileMixin:
    tabs = [
        {"name":"Details", "url":"users_my_profile_overview", "requires_pk":False},
        {"name":"Change password", "url":"users_my_profile_change_password", "requires_pk":False},
        {"name":"List view preferences", "url":"users_my_profile_list_view_preference","requires_pk":False},
        {"name":"Detail view preferences", "url":"users_my_profile_detail_view_preference","requires_pk":False}
    ]
    exclude_header = True

    extra_context = {"disable_tab_select":True}

    def get_object(self):
        return User.objects.get(pk=self.request.user.pk)
    model = User

router = BloomerpRouter()

@router.bloomerp_route(
    path="my-profile/",
    models=User,
    route_type="list",
    name="Profile",
    description="Overview of profile",
    url_name="my_profile_overview"
)
class BloomerpProfileView(BloomerpModelFormViewMixin, ProfileMixin, BloomerpBaseDetailView):
    template_name = 'auth_views/profile_overview.html'
    fields = ['first_name', 'last_name', 'date_view_preference', 'datetime_view_preference', 'avatar']
    success_url = reverse_lazy('profile_overview')

@router.bloomerp_route(
    path="my-profile/reset-password/",
    models=User,
    route_type="list",
    name="Reset password",
    description="Reset password for a user",
    url_name="my_profile_change_password"
)
class BloomerpProfilePasswordResetView(ProfileMixin, BloomerpBaseDetailView, FormView):
    template_name = 'auth_views/profile_password_reset.html'
    form_class = PasswordChangeForm
    success_url = reverse_lazy('profile_overview')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.request.user
        return kwargs

    def form_valid(self, form):
        self.object = self.get_object()
        return super().form_valid(form)    
    
    def form_invalid(self, form):
        self.object = self.get_object()
        return super().form_invalid(form)
    

@router.bloomerp_route(
    path="my-profile/list-view-preference/",
    models=User,
    route_type="list",
    name="List view preference",
    description="List view preference for a user",
    url_name="my_profile_list_view_preference"
)
class UserListViewPreferenceView(ProfileMixin, BloomerpBaseDetailView):
    template_name = 'user_views/list_view_preference_view.html'
    permission_required = ['users.change_userlistviewpreference']
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        content_type_list = ContentType.objects.all()
        content_type_id = self.request.GET.get('content_type_id',None)

        
        if content_type_id:
            context['selected_model'] = ContentType.objects.get(pk=int(content_type_id)).model

            form = ListViewFieldsSelectForm(
                content_type=ContentType.objects.get(pk=content_type_id),
                user=self.request.user)
            
            context['form'] = form

        context['content_type_list'] = content_type_list
        return context

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        content_type_id = request.GET.get('content_type_id',None)
        return_to = request.GET.get('return_to',None)

        form = ListViewFieldsSelectForm(
                content_type=ContentType.objects.get(pk=content_type_id),
                user=self.request.user,
                data=request.POST
            )
        
        if form.is_valid():
            form.save()
            messages.success(self.request, 'Preferences saved successfully.')
            if return_to:
                return redirect(return_to)
            else:
                return redirect(reverse('list_view_preference') + f'?content_type_id={content_type_id}')
        else:
            return render(request, self.template_name, {'form': form})

        
@router.bloomerp_route(
    path="my-profile/detail-view-preferences/",
    models=User,
    route_type="list",
    name="Detail view preference",
    description="Detail view preference for a user",
    url_name="my_profile_detail_view_preference"
)
class UserDetailViewPreferenceView(ProfileMixin, BloomerpBaseDetailView):
    template_name = 'user_views/bloomerp_detail_view_preference_view.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        content_type_id = self.request.GET.get('content_type_id',None)

        list = UserDetailViewPreference.get_application_fields_info(content_type_id,self.request.user)

        Formset = formset_factory(UserDetailViewPreferenceForm, formset=UserDetailViewPreferenceFormset,extra=0)
        
        formset = Formset(initial=list)

        context['formset'] = formset
        context['content_type_list'] = ContentType.objects.all()
        if content_type_id:
            context['selected_model'] = ContentType.objects.get(pk=int(content_type_id)).model

        return context

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        content_type_id = request.GET.get('content_type_id',None)
        
        return_to = request.GET.get('return_to',None)

        Formset = formset_factory(UserDetailViewPreferenceForm, formset=UserDetailViewPreferenceFormset)


        formset = Formset(request.POST)

        if formset.is_valid():
            formset.user = self.request.user
            formset.save()
            messages.success(self.request, 'Preferences saved successfully.')
            if return_to:
                return redirect(return_to)
            else:
                return redirect(reverse('detail_view_preference') + f'?content_type_id={content_type_id}')
        else:
            return render(request, self.template_name, {'formset': formset})


from bloomerp.forms.auth import BloomerpUserCreationForm
@router.bloomerp_route(
    path="create",
    name="Create user",
    url_name="add",
    description="Create a new object from User",
    route_type="list",
    models=User
)
class UserCreateView(
        PermissionRequiredMixin, 
        SuccessMessageMixin,
        HtmxMixin,
        FormView):
    template_name = "create_views/bloomerp_create_view.html"
    fields = None
    model = None
    exclude = []
    success_message = "Object was created successfully."
    form_class = BloomerpUserCreationForm
    success_url = reverse_lazy("users_list")


    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["model_name"] = self.model._meta.verbose_name
        context["model_name_plural"] = self.model._meta.verbose_name_plural
        context["list_view_url"] = model_name_plural_underline(self.model) + "_list"
        return context

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.add_{self.model._meta.model_name}"]

    def get_success_message(self, cleaned_data):
        return f"User was created successfully."

    


from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.utils.llm import BloomerpOpenAI
from bloomerp.models import ApplicationField
from django.contrib.auth.decorators import login_required

@login_required
@route('llm_executor')
def llm_executor(request:HttpRequest) -> HttpResponse:
    '''
    Component to execute LLM queries.
    '''

    query_type = request.GET.get('llm_query_type', 'sql')
    query = request.GET.get('llm_query', None)

    llm_query_types = ['sql', 'document_template']

    if not query:
        return HttpResponse('No llm query provided')

    if query_type not in llm_query_types:
        return HttpResponse('Invalid llm query type, must be one of: ' + ', '.join(llm_query_types))

    # Init the OpenAI class
    openai = BloomerpOpenAI()

    # Check if the key is valid
    if not openai.is_valid_key():
        return HttpResponse('Invalid OpenAI key')


    if query_type == 'sql':
        db_tables_and_columns = ApplicationField.get_db_tables_and_columns()
        sql_query = openai.create_sql_query(query, db_tables_and_columns)
        return HttpResponse(sql_query)
    
    elif query_type == 'document_template':
        
        return HttpResponse('Document template AI is not implemented yet')

    context = {}
    return render(request, 'components/llm_executor.html', context)


from setuptools import find_packages, setup

setup(
    name="Bloomerp",
    version="0.1.2",
    packages=find_packages(),
    include_package_data=True,
    license="MIT License",
    description="Bloomerp is an open source Business Management Software framework that let's you create a fully functioning business management applications by just Django models.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/davidbloomer11/Bloomerp",
    author="David Bloomer",
    author_email="bloomer.david@outlook.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'django>=4.2',
        'requests>=2.0.0',
        'djangorestframework>=3.12.0',
        'django-htmx>=1.0.0',
        'django-formtools>=2.3',
        'Pillow>=8.0.0',
        'psycopg2==2.9.9',
        'django-crispy-forms>=2.2',
        'django-filter>=24.3',
        'xhtml2pdf>=0.2.16',
        'pandas>=2.0.0',
        'openpyxl>=3.1.5',
        'openai>=1.54.1',
        'plotly>=5.0.0'
    ],
    extras_require={
        'images': ['Pillow>=8.0.0'],
    },
)